"""Authentication helpers for Dewey API and browser UI."""

from __future__ import annotations

import base64
import json
import secrets
from typing import Any

from daylily_cognito import build_authorization_url, exchange_authorization_code
from fastapi import Depends, HTTPException, Request, status
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer

from dewey_service.rbac import Role, normalize_session_profile, profile_has_role
from dewey_service.settings import Settings


class AuthError(RuntimeError):
    """Raised when authentication flow fails."""


def decode_jwt_claims_noverify(token: str) -> dict[str, Any]:
    if not token or "." not in token:
        return {}
    try:
        parts = token.split(".")
        if len(parts) < 2:
            return {}
        payload = parts[1]
        payload += "=" * ((4 - len(payload) % 4) % 4)
        decoded = base64.urlsafe_b64decode(payload.encode("utf-8")).decode("utf-8")
        parsed = json.loads(decoded)
        return parsed if isinstance(parsed, dict) else {}
    except Exception:
        return {}


def generate_state() -> str:
    return secrets.token_urlsafe(24)


def build_cognito_login_url(*, settings: Settings, state: str) -> str:
    domain = str(settings.cognito_domain or "").strip().rstrip("/")
    if domain.startswith("https://"):
        domain = domain[len("https://") :]
    return build_authorization_url(
        domain=domain,
        client_id=settings.cognito_app_client_id,
        redirect_uri=settings.cognito_redirect_uri,
        state=state,
    )


def build_cognito_logout_url(*, settings: Settings, state: str | None = None) -> str:
    import urllib.parse
    domain = str(settings.cognito_domain or "").strip().rstrip("/")
    if domain.startswith("https://"):
        domain = domain[len("https://") :]
    logout_target = settings.cognito_logout_url
    query: dict[str, str] = {
        "client_id": settings.cognito_app_client_id,
        "logout_uri": logout_target.rstrip("/"),
    }
    if state:
        query["state"] = state
    params = urllib.parse.urlencode(query)
    return f"https://{domain}/logout?{params}"


def exchange_code(*, settings: Settings, code: str) -> dict[str, Any]:
    domain = str(settings.cognito_domain or "").strip().rstrip("/")
    if domain.startswith("https://"):
        domain = domain[len("https://") :]
    try:
        return exchange_authorization_code(
            domain=domain,
            client_id=settings.cognito_app_client_id,
            code=code,
            redirect_uri=settings.cognito_redirect_uri,
            client_secret=settings.cognito_app_client_secret or None,
        )
    except RuntimeError as exc:
        raise AuthError(str(exc)) from exc


def require_api_auth(settings: Settings):
    bearer = HTTPBearer(auto_error=False)

    def _require_api_auth(
        credentials: HTTPAuthorizationCredentials | None = Depends(bearer),
    ) -> str:
        if credentials is None:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Missing bearer token",
                headers={"WWW-Authenticate": "Bearer"},
            )
        token = str(credentials.credentials or "").strip()
        if token not in settings.api_tokens():
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid bearer token",
                headers={"WWW-Authenticate": "Bearer"},
            )
        return token

    return _require_api_auth


def require_ui_session(request: Request) -> dict[str, Any]:
    profile = request.session.get("operator_profile")
    if not isinstance(profile, dict):
        store = getattr(request.app.state, "observability", None)
        if store is not None:
            store.record_auth_event(
                status="denied",
                mode="anonymous",
                detail="ui_session",
                service_principal=False,
            )
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Login required")
    store = getattr(request.app.state, "observability", None)
    if store is not None:
        store.record_auth_event(
            status="ok",
            mode="cognito",
            detail="ui_session",
            service_principal=False,
        )
    request.state.auth_mode = "cognito"
    if "roles" not in profile:
        return normalize_session_profile(
            email=profile.get("email"),
            sub=profile.get("sub"),
            groups=profile.get("groups"),
        )
    return profile


def require_ui_admin_session(request: Request) -> dict[str, Any]:
    profile = require_ui_session(request)
    if not profile_has_role(profile, Role.ADMIN):
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Admin access required",
        )
    return profile


def build_session_profile(
    *,
    settings: Settings,
    email: str,
    sub: str,
    groups: list[str] | tuple[str, ...] | None,
) -> dict[str, Any]:
    return normalize_session_profile(
        email=email,
        sub=sub,
        groups=groups,
        group_role_map=settings.cognito_group_role_map,
    )


def require_observability_access(settings: Settings):
    bearer = HTTPBearer(auto_error=False)

    def _require_observability_access(
        request: Request,
        credentials: HTTPAuthorizationCredentials | None = Depends(bearer),
    ) -> dict[str, Any]:
        profile = request.session.get("operator_profile")
        if isinstance(profile, dict):
            store = getattr(request.app.state, "observability", None)
            if store is not None:
                store.record_auth_event(
                    status="ok",
                    mode="cognito",
                    detail="session",
                    service_principal=False,
                )
            request.state.auth_mode = "cognito"
            return {"auth_mode": "cognito", "service_principal": False, "profile": profile}

        if credentials is not None:
            token = str(credentials.credentials or "").strip()
            if token in settings.api_tokens():
                store = getattr(request.app.state, "observability", None)
                if store is not None:
                    store.record_auth_event(
                        status="ok",
                        mode="service_token",
                        detail="bearer",
                        service_principal=True,
                    )
                request.state.auth_mode = "service_token"
                return {"auth_mode": "service_token", "service_principal": True}

        store = getattr(request.app.state, "observability", None)
        if store is not None:
            store.record_auth_event(
                status="denied",
                mode="anonymous",
                detail="observability",
                service_principal=False,
            )
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Missing bearer token",
            headers={"WWW-Authenticate": "Bearer"},
        )

    return _require_observability_access



"""Storage helpers for Dewey artifact lifecycle operations."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any


class StorageError(RuntimeError):
    """Base storage operation failure."""


class StorageObjectNotFoundError(StorageError):
    """Raised when a referenced storage object does not exist."""


class StoragePermissionError(StorageError):
    """Raised when Dewey lacks permission for a storage operation."""


@dataclass(frozen=True)
class StorageObject:
    bucket: str
    key: str
    version_id: str | None = None
    size: int | None = None
    content_type: str | None = None
    storage_class: str | None = None
    etag: str | None = None


class S3StorageClient:
    """Small S3 adapter kept intentionally narrow for Dewey flows."""

    def __init__(self, *, profile: str | None = None, region: str | None = None) -> None:
        try:
            import boto3
            from botocore.exceptions import ClientError
        except ImportError as exc:  # pragma: no cover - runtime guard
            raise RuntimeError("boto3 is required for Dewey storage operations") from exc

        session_kwargs: dict[str, Any] = {}
        if str(profile or "").strip():
            session_kwargs["profile_name"] = str(profile).strip()
        if str(region or "").strip():
            session_kwargs["region_name"] = str(region).strip()

        session = boto3.session.Session(**session_kwargs)
        self._client = session.client("s3")
        self._client_error = ClientError

    def head_object(
        self,
        *,
        bucket: str,
        key: str,
        version_id: str | None = None,
    ) -> StorageObject:
        params: dict[str, Any] = {"Bucket": bucket, "Key": key}
        if version_id:
            params["VersionId"] = version_id
        try:
            response = self._client.head_object(**params)
        except self._client_error as exc:
            raise self._translate_error(exc, bucket=bucket, key=key) from exc
        return self._to_storage_object(bucket=bucket, key=key, response=response)

    def copy_object(
        self,
        *,
        source_bucket: str,
        source_key: str,
        dest_bucket: str,
        dest_key: str,
    ) -> StorageObject:
        try:
            self._client.copy_object(
                Bucket=dest_bucket,
                Key=dest_key,
                CopySource={"Bucket": source_bucket, "Key": source_key},
            )
        except self._client_error as exc:
            raise self._translate_error(exc, bucket=source_bucket, key=source_key) from exc
        return self.head_object(bucket=dest_bucket, key=dest_key)

    def put_bytes(
        self,
        *,
        bucket: str,
        key: str,
        body: bytes,
        content_type: str | None = None,
    ) -> StorageObject:
        params: dict[str, Any] = {
            "Bucket": bucket,
            "Key": key,
            "Body": body,
        }
        if content_type:
            params["ContentType"] = content_type
        try:
            self._client.put_object(**params)
        except self._client_error as exc:
            raise self._translate_error(exc, bucket=bucket, key=key) from exc
        return self.head_object(bucket=bucket, key=key)

    def put_object_tags(self, *, bucket: str, key: str, tags: dict[str, str]) -> None:
        merged = dict(self.get_object_tags(bucket=bucket, key=key))
        merged.update({str(k): str(v) for k, v in tags.items() if str(v).strip()})
        try:
            self._client.put_object_tagging(
                Bucket=bucket,
                Key=key,
                Tagging={
                    "TagSet": [
                        {"Key": tag_key, "Value": tag_value}
                        for tag_key, tag_value in sorted(merged.items())
                    ]
                },
            )
        except self._client_error as exc:
            raise self._translate_error(exc, bucket=bucket, key=key) from exc

    def get_object_tags(self, *, bucket: str, key: str) -> dict[str, str]:
        try:
            response = self._client.get_object_tagging(Bucket=bucket, Key=key)
        except self._client_error as exc:
            raise self._translate_error(exc, bucket=bucket, key=key) from exc
        tags: dict[str, str] = {}
        for item in response.get("TagSet", []):
            tag_key = str(item.get("Key") or "").strip()
            if tag_key:
                tags[tag_key] = str(item.get("Value") or "")
        return tags

    def set_retention(
        self,
        *,
        bucket: str,
        key: str,
        mode: str,
        retain_until: datetime,
    ) -> None:
        try:
            self._client.put_object_retention(
                Bucket=bucket,
                Key=key,
                Retention={
                    "Mode": str(mode or "GOVERNANCE").strip().upper(),
                    "RetainUntilDate": retain_until,
                },
            )
        except self._client_error as exc:
            raise self._translate_error(exc, bucket=bucket, key=key) from exc

    def generate_presigned_get_url(
        self,
        *,
        bucket: str,
        key: str,
        expires_in: int,
        version_id: str | None = None,
    ) -> str:
        params: dict[str, Any] = {"Bucket": bucket, "Key": key}
        if version_id:
            params["VersionId"] = version_id
        try:
            return str(
                self._client.generate_presigned_url(
                    "get_object",
                    Params=params,
                    ExpiresIn=max(60, int(expires_in)),
                )
            )
        except self._client_error as exc:
            raise self._translate_error(exc, bucket=bucket, key=key) from exc

    def generate_presigned_upload(
        self,
        *,
        bucket: str,
        key: str,
        expires_in: int,
        content_type: str | None = None,
    ) -> dict[str, Any]:
        params: dict[str, Any] = {"Bucket": bucket, "Key": key}
        headers: dict[str, str] = {}
        if content_type:
            params["ContentType"] = content_type
            headers["Content-Type"] = content_type
        try:
            url = self._client.generate_presigned_url(
                "put_object",
                Params=params,
                ExpiresIn=max(60, int(expires_in)),
            )
        except self._client_error as exc:
            raise self._translate_error(exc, bucket=bucket, key=key) from exc
        return {
            "method": "PUT",
            "url": str(url),
            "headers": headers,
        }

    def _to_storage_object(
        self,
        *,
        bucket: str,
        key: str,
        response: dict[str, Any],
    ) -> StorageObject:
        return StorageObject(
            bucket=bucket,
            key=key,
            version_id=response.get("VersionId"),
            size=response.get("ContentLength"),
            content_type=response.get("ContentType"),
            storage_class=response.get("StorageClass"),
            etag=str(response.get("ETag") or "").strip('"') or None,
        )

    def _translate_error(self, exc: Exception, *, bucket: str, key: str) -> StorageError:
        error = getattr(exc, "response", {}).get("Error", {})
        code = str(error.get("Code") or "").strip()
        message = str(error.get("Message") or f"{bucket}/{key}").strip()
        if code in {"404", "NoSuchKey", "NotFound"}:
            return StorageObjectNotFoundError(message)
        if code in {"403", "AccessDenied"}:
            return StoragePermissionError(message)
        return StorageError(message)



from __future__ import annotations


def test_dewey_allows_approved_origin_preflight(client) -> None:
    response = client.options(
        "/ui",
        headers={
            "Origin": "https://portal.lsmc.bio",
            "Access-Control-Request-Method": "GET",
        },
    )

    assert response.status_code == 200
    assert response.headers["access-control-allow-origin"] == "https://portal.lsmc.bio"


def test_dewey_rejects_disallowed_origin(client) -> None:
    response = client.options(
        "/ui",
        headers={
            "Origin": "https://evil.example.com",
            "Access-Control-Request-Method": "GET",
        },
    )

    assert response.status_code == 403
    assert response.text == "Origin not allowed"


def test_dewey_rejects_disallowed_host(client) -> None:
    response = client.get("/ui", headers={"host": "evil.example.com"}, follow_redirects=False)

    assert response.status_code == 400



from __future__ import annotations

import importlib.metadata
from pathlib import Path

import pytest
from cli_core_yo import output
from cli_core_yo.app import run as run_cli

import dewey_service.cli as cli_module
import dewey_service.cli.server as server_cli
from dewey_service.settings import Settings


def _invoke(argv: list[str]) -> int:
    output._reset_console()
    return run_cli(cli_module.spec, argv)


def test_cli_help_renders(capsys) -> None:
    exit_code = _invoke(["--help"])
    captured = capsys.readouterr()

    assert exit_code == 0
    assert "Usage:" in captured.out
    assert "version" in captured.out
    assert "info" in captured.out
    assert "config" in captured.out
    assert "env" in captured.out
    assert "server" in captured.out


def test_cli_version_renders(capsys) -> None:
    exit_code = _invoke(["version"])
    captured = capsys.readouterr()

    assert exit_code == 0
    assert importlib.metadata.version("dewey-service") in captured.out


def test_cli_info_renders(monkeypatch, capsys) -> None:
    monkeypatch.setattr(
        cli_module,
        "get_settings",
        lambda: Settings(
            cognito_domain="https://auth.example.com",
            cognito_app_client_id="client-1",
            cognito_redirect_uri="https://localhost:8914/auth/callback",
            cognito_logout_url="https://localhost:8914/login",
        ),
    )

    exit_code = _invoke(["info"])
    captured = capsys.readouterr()

    assert exit_code == 0
    assert "Dewey Info" in captured.out
    assert "Database Backend" in captured.out
    assert "TapDB Client" in captured.out


def test_server_status_reports_not_running(monkeypatch, capsys) -> None:
    monkeypatch.setattr(server_cli, "read_pid", lambda _pid_file: None)

    exit_code = _invoke(["server", "status"])
    captured = capsys.readouterr()

    assert exit_code == 0
    assert "not running" in captured.out


def test_server_restart_uses_background_start(monkeypatch, capsys) -> None:
    calls: list[tuple[str, object]] = []

    monkeypatch.setattr(
        server_cli,
        "_stop_server",
        lambda: calls.append(("stop", None)),
    )
    monkeypatch.setattr(
        server_cli,
        "_start_server",
        lambda **kwargs: calls.append(("start", kwargs)),
    )
    monkeypatch.setattr(server_cli, "_validate_cognito_uris_for_port", lambda **kwargs: None)
    monkeypatch.setattr(server_cli.time, "sleep", lambda _seconds: None)

    exit_code = _invoke(["server", "restart"])
    capsys.readouterr()

    assert exit_code == 0
    assert calls == [
        ("stop", None),
        (
            "start",
            {
                "host": "0.0.0.0",
                "port": 8914,
                "reload": False,
                "background": True,
            },
        ),
    ]


def test_config_init_show_validate_and_status(monkeypatch, tmp_path: Path, capsys) -> None:
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    monkeypatch.setenv("DEWEY_DEPLOYMENT_CODE", "local")

    init_exit = _invoke(["config", "init"])
    init_output = capsys.readouterr().out
    config_path = tmp_path / "dewey-local" / "dewey-config-local.yaml"

    assert init_exit == 0
    assert "Config file created" in init_output
    assert config_path.exists()

    show_exit = _invoke(["config", "show"])
    show_output = capsys.readouterr().out
    assert show_exit == 0
    assert "application:" in show_output
    assert "database:" in show_output


def test_cli_requires_hyphenated_conda_env(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("CONDA_DEFAULT_ENV", "DEWEY")
    with pytest.raises(SystemExit, match="deployment-scoped conda environment name with '-'"):
        cli_module._enforce_conda_env_contract(["server", "status"])


def test_cli_requires_active_conda_env(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("CONDA_DEFAULT_ENV", raising=False)
    with pytest.raises(
        SystemExit, match="requires an active deployment-scoped conda environment"
    ):
        cli_module._enforce_conda_env_contract(["server", "status"])


def test_cli_accepts_hyphenated_conda_env(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("CONDA_DEFAULT_ENV", "DEWEY-local2")
    cli_module._enforce_conda_env_contract(["server", "status"])


def test_cli_skip_conda_env_check_flag_is_stripped() -> None:
    args, skip = cli_module._strip_skip_conda_env_check_flag(
        ["--skip-conda-env-check", "server", "status"]
    )
    assert skip is True
    assert args == ["server", "status"]


def test_config_validate_and_status(monkeypatch, tmp_path: Path, capsys) -> None:
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    monkeypatch.setenv("DEWEY_DEPLOYMENT_CODE", "local")

    init_exit = _invoke(["config", "init"])
    capsys.readouterr()
    assert init_exit == 0

    validate_exit = _invoke(["config", "validate"])
    validate_output = capsys.readouterr().out
    assert validate_exit == 0
    assert "Config is valid" in validate_output

    status_exit = _invoke(["config", "status"])
    status_output = capsys.readouterr().out
    assert status_exit == 0
    assert "Config path:" in status_output
    assert "dewey-config-local.yaml" in status_output
    assert '"tapdb_database_name": "dewey"' in status_output


def test_env_commands_render(monkeypatch, capsys) -> None:
    monkeypatch.setenv("DEWEY_ACTIVE", "1")
    monkeypatch.setenv("DEWEY_PROJECT_ROOT", "/tmp/dewey")

    status_exit = _invoke(["env", "status"])
    status_output = capsys.readouterr().out
    assert status_exit == 0
    assert "active" in status_output
    assert "/tmp/dewey" in status_output

    activate_exit = _invoke(["env", "activate"])
    activate_output = capsys.readouterr().out
    assert activate_exit == 0
    assert f"source {cli_module.ACTIVATE_SCRIPT}" in activate_output

    deactivate_exit = _invoke(["env", "deactivate"])
    deactivate_output = capsys.readouterr().out
    assert deactivate_exit == 0
    assert f"source {cli_module.DEACTIVATE_SCRIPT}" in deactivate_output

    reset_exit = _invoke(["env", "reset"])
    reset_output = capsys.readouterr().out
    assert reset_exit == 0
    assert f"source {cli_module.DEACTIVATE_SCRIPT}" in reset_output
    assert f"source {cli_module.ACTIVATE_SCRIPT}" in reset_output



"""Seed the Dewey TapDB JSON template pack through TapDB's loader."""

from __future__ import annotations

from pathlib import Path

from daylily_tapdb import resolve_seed_config_dirs, seed_templates, validate_template_configs

from dewey_service.tapdb_backend import TapDBBackend


def main() -> None:
    backend = TapDBBackend(app_username="dewey")
    config_root = Path(__file__).resolve().parents[1] / "config" / "tapdb_templates"
    config_dirs = resolve_seed_config_dirs(config_root)
    templates, issues = validate_template_configs(config_dirs, strict=True)
    errors = [issue for issue in issues if issue.level == "error"]
    if errors:
        joined = "; ".join(issue.message for issue in errors)
        raise RuntimeError(f"Dewey template pack validation failed: {joined}")
    with backend.session_scope(commit=True) as session:
        seed_templates(session, templates, overwrite=True)


if __name__ == "__main__":
    main()



from __future__ import annotations


def test_register_artifact_idempotent_replay(client) -> None:
    headers = {"Authorization": "Bearer token-123", "Idempotency-Key": "idem-replay-1"}
    payload = {
        "artifact_type": "fastq",
        "storage_backend": "s3",
        "bucket": "bucket-1",
        "key": "reads/r1.fastq.gz",
    }
    first = client.post("/api/v1/artifacts", headers=headers, json=payload)
    second = client.post("/api/v1/artifacts", headers=headers, json=payload)
    assert first.status_code == 200
    assert second.status_code == 200
    assert first.json()["artifact_euid"] == second.json()["artifact_euid"]


def test_register_artifact_idempotency_conflict(client) -> None:
    headers = {"Authorization": "Bearer token-123", "Idempotency-Key": "idem-conflict-1"}
    payload_a = {
        "artifact_type": "fastq",
        "storage_backend": "s3",
        "bucket": "bucket-1",
        "key": "reads/a.fastq.gz",
    }
    payload_b = {
        "artifact_type": "fastq",
        "storage_backend": "s3",
        "bucket": "bucket-1",
        "key": "reads/b.fastq.gz",
    }
    first = client.post("/api/v1/artifacts", headers=headers, json=payload_a)
    second = client.post("/api/v1/artifacts", headers=headers, json=payload_b)
    assert first.status_code == 200
    assert second.status_code == 409
    assert "Idempotency-Key reuse" in second.json()["detail"]


def test_mutation_requires_idempotency_key(client) -> None:
    response = client.post(
        "/api/v1/artifact-sets",
        headers={"Authorization": "Bearer token-123"},
        json={"artifact_set_type": "bundle"},
    )
    assert response.status_code == 400
    assert "Idempotency-Key" in response.json()["detail"]



"""TapDB backend wiring for Dewey artifact persistence."""

from __future__ import annotations

import datetime as dt
import inspect
import os
from contextlib import contextmanager
from hashlib import sha256
from time import monotonic
from typing import Any, Generator, Optional, cast

from daylily_tapdb import (
    InstanceFactory,
    TAPDBConnection,
    TemplateManager,
    generic_instance,
    generic_instance_lineage,
)
from daylily_tapdb.cli.context import resolve_context
from daylily_tapdb.cli.db_config import get_db_config_for_env
from sqlalchemy import and_
from sqlalchemy.orm import Session


def utc_now_iso() -> str:
    return dt.datetime.now(dt.timezone.utc).isoformat().replace("+00:00", "Z")


def sha256_json(value: Any) -> str:
    payload = str(value).encode("utf-8")
    return sha256(payload).hexdigest()


def _parse_template_code(template_code: str) -> tuple[str, str, str, str]:
    parts = template_code.strip("/").split("/")
    if len(parts) != 4:
        raise ValueError(f"Invalid template code: {template_code}")
    return parts[0], parts[1], parts[2], parts[3]


ARTIFACT_TEMPLATE = "dewey/data/artifact/1.0/"
ARTIFACT_SET_TEMPLATE = "dewey/data/artifact_set/1.0/"
SHARE_REFERENCE_TEMPLATE = "dewey/data/share_reference/1.0/"
EXTERNAL_OBJECT_TEMPLATE = "dewey/integration/external_object/1.0/"
EXTERNAL_OBJECT_RELATION_TEMPLATE = "dewey/integration/external_object_relation/1.0/"
LITERATURE_SAVE_TEMPLATE = "dewey/access/literature_save/1.0/"
ANOMALY_TEMPLATE = "dewey/operational/anomaly/1.0/"
IDEMPOTENCY_TEMPLATE = "dewey/system/idempotency_request/1.0/"


TEMPLATE_DEFINITIONS: tuple[str, ...] = (
    ARTIFACT_TEMPLATE,
    ARTIFACT_SET_TEMPLATE,
    SHARE_REFERENCE_TEMPLATE,
    EXTERNAL_OBJECT_TEMPLATE,
    EXTERNAL_OBJECT_RELATION_TEMPLATE,
    LITERATURE_SAVE_TEMPLATE,
    ANOMALY_TEMPLATE,
    IDEMPOTENCY_TEMPLATE,
)


class TapDBBackend:
    """TapDB-backed repository for Dewey domain entities."""

    def __init__(self, app_username: str = "dewey"):
        os.environ.setdefault("TAPDB_STRICT_NAMESPACE", "1")
        env = (os.environ.get("TAPDB_ENV") or "").strip()
        if not env:
            raise RuntimeError("TAPDB_ENV is required (dev|test|prod)")

        try:
            resolve_context(require_keys=True)
            cfg = get_db_config_for_env(env)
        except Exception as exc:
            raise RuntimeError(
                "TapDB is not configured for Dewey.\n"
                "Required env: TAPDB_CLIENT_ID, TAPDB_DATABASE_NAME, TAPDB_ENV"
            ) from exc

        db_hostname = f"{cfg['host']}:{cfg['port']}"
        engine_type = (cfg.get("engine_type") or "local").strip().lower()
        region = (cfg.get("region") or os.environ.get("AWS_REGION") or "us-west-2").strip()
        iam_auth_raw = str(cfg.get("iam_auth") or "").strip().lower()
        iam_auth = iam_auth_raw in {"1", "true", "yes", "on"}
        secret_arn = cfg.get("secret_arn") or cfg.get("master_secret_arn")

        self.connection = TAPDBConnection(
            db_hostname=db_hostname,
            db_user=cfg["user"],
            db_pass=cfg["password"],
            db_name=cfg["database"],
            app_username=app_username,
            engine_type=engine_type if engine_type != "local" else None,
            region=region,
            iam_auth=iam_auth,
            secret_arn=secret_arn,
        )
        self.templates = TemplateManager()
        self.factory = InstanceFactory(self.templates)
        self.observability = None

    @contextmanager
    def session_scope(self, commit: bool = False) -> Generator[Session, None, None]:
        started = monotonic()
        success = False
        try:
            with self.connection.session_scope(commit=commit) as session:
                yield session
                success = True
        finally:
            if self.observability is not None:
                self.observability.record_db_operation(
                    label=self._operation_label(),
                    duration_ms=(monotonic() - started) * 1000,
                    success=success,
                )

    def _operation_label(self) -> str:
        for frame_info in inspect.stack(context=0):
            filename = str(frame_info.filename or "")
            if filename.endswith("/dewey_service/service.py"):
                return f"service:{frame_info.function}"
            if filename.endswith("/dewey_service/app.py"):
                return f"app:{frame_info.function}"
        return "tapdb_session"

    def ensure_templates(self, session: Session) -> None:
        missing = [
            template_code
            for template_code in TEMPLATE_DEFINITIONS
            if self.templates.get_template(session, template_code) is None
        ]
        if missing:
            joined = ", ".join(missing)
            raise RuntimeError(
                "Missing Dewey templates. Seed the Dewey TapDB JSON pack before "
                f"running the service: {joined}"
            )

    def create_instance(
        self,
        session: Session,
        *,
        template_code: str,
        name: str,
        json_addl: dict[str, Any],
        status: str = "active",
    ) -> generic_instance:
        template = self.templates.get_template(session, template_code)
        if template is None:
            raise RuntimeError(f"Missing template: {template_code}")

        instance = self.factory.create_instance(
            session=session,
            template_code=template_code,
            name=name,
            properties={},
            create_children=False,
        )
        instance.json_addl = dict(json_addl)
        instance.bstatus = status
        instance.is_singleton = False
        session.flush()
        return instance

    def update_instance_json(
        self, session: Session, instance: generic_instance, updates: dict[str, Any]
    ) -> None:
        payload = dict(instance.json_addl or {})
        payload.update(updates)
        instance.json_addl = payload
        session.flush()

    def _template_query(
        self,
        session: Session,
        *,
        template_code: str,
        for_update: bool = False,
    ):
        template = self.templates.get_template(session, template_code)
        if template is None:
            return None
        query = session.query(generic_instance).filter(
            generic_instance.template_uid == template.uid,
            generic_instance.is_deleted.is_(False),
        )
        if for_update:
            query = query.with_for_update()
        return query

    def find_by_euid(
        self,
        session: Session,
        *,
        template_code: str,
        euid: str,
        for_update: bool = False,
    ) -> Optional[generic_instance]:
        if not euid:
            return None
        query = self._template_query(session, template_code=template_code, for_update=for_update)
        if query is None:
            return None
        return query.filter(generic_instance.euid == euid).first()

    def find_by_json_field(
        self,
        session: Session,
        *,
        template_code: str,
        field: str,
        value: str,
    ) -> Optional[generic_instance]:
        query = self._template_query(session, template_code=template_code)
        if query is None:
            return None
        return query.filter(generic_instance.json_addl[field].as_string() == value).first()

    def list_by_template(
        self,
        session: Session,
        *,
        template_code: str,
        limit: int = 200,
    ) -> list[generic_instance]:
        query = self._template_query(session, template_code=template_code)
        if query is None:
            return []
        return cast(
            list[generic_instance],
            query.order_by(generic_instance.created_dt.desc()).limit(limit).all(),
        )

    def create_lineage(
        self,
        session: Session,
        *,
        parent: generic_instance,
        child: generic_instance,
        relationship_type: str,
        name: str | None = None,
    ) -> generic_instance_lineage:
        existing = (
            session.query(generic_instance_lineage)
            .filter(
                generic_instance_lineage.parent_instance_uid == parent.uid,
                generic_instance_lineage.child_instance_uid == child.uid,
                generic_instance_lineage.relationship_type == relationship_type,
                generic_instance_lineage.is_deleted.is_(False),
            )
            .first()
        )
        if existing is not None:
            return existing

        lineage = generic_instance_lineage(
            name=name or f"{parent.euid}->{child.euid}:{relationship_type}",
            polymorphic_discriminator="generic_instance_lineage",
            category="generic",
            type="lineage",
            subtype="instance_lineage",
            version="1.0.0",
            bstatus="active",
            json_addl={},
            is_singleton=False,
            parent_type=parent.polymorphic_discriminator,
            child_type=child.polymorphic_discriminator,
            relationship_type=relationship_type,
            parent_instance_uid=parent.uid,
            child_instance_uid=child.uid,
        )
        session.add(lineage)
        session.flush()
        return lineage

    def delete_lineage(
        self,
        session: Session,
        *,
        parent: generic_instance,
        child: generic_instance,
        relationship_type: str,
    ) -> bool:
        lineage = (
            session.query(generic_instance_lineage)
            .filter(
                generic_instance_lineage.parent_instance_uid == parent.uid,
                generic_instance_lineage.child_instance_uid == child.uid,
                generic_instance_lineage.relationship_type == relationship_type,
                generic_instance_lineage.is_deleted.is_(False),
            )
            .first()
        )
        if lineage is None:
            return False
        lineage.is_deleted = True
        lineage.bstatus = "deleted"
        session.flush()
        return True

    def list_children(
        self,
        session: Session,
        *,
        parent: generic_instance,
        relationship_type: str | None = None,
    ) -> list[generic_instance]:
        query = (
            session.query(generic_instance)
            .join(
                generic_instance_lineage,
                generic_instance_lineage.child_instance_uid == generic_instance.uid,
            )
            .filter(
                generic_instance_lineage.parent_instance_uid == parent.uid,
                generic_instance_lineage.is_deleted.is_(False),
                generic_instance.is_deleted.is_(False),
            )
        )
        if relationship_type:
            query = query.filter(generic_instance_lineage.relationship_type == relationship_type)
        return cast(list[generic_instance], query.all())

    def list_parents(
        self,
        session: Session,
        *,
        child: generic_instance,
        relationship_type: str | None = None,
    ) -> list[generic_instance]:
        query = (
            session.query(generic_instance)
            .join(
                generic_instance_lineage,
                generic_instance_lineage.parent_instance_uid == generic_instance.uid,
            )
            .filter(
                generic_instance_lineage.child_instance_uid == child.uid,
                generic_instance_lineage.is_deleted.is_(False),
                generic_instance.is_deleted.is_(False),
            )
        )
        if relationship_type:
            query = query.filter(generic_instance_lineage.relationship_type == relationship_type)
        return cast(list[generic_instance], query.all())

    def find_lineage_instance(
        self,
        session: Session,
        *,
        source: generic_instance,
        relation_template_code: str,
        parent_relationship_type: str,
        child_relationship_type: str,
    ) -> Optional[generic_instance]:
        # Resolve relation instances by traversing lineages through the relation object.
        relation_template = self.templates.get_template(session, relation_template_code)
        if relation_template is None:
            return None

        candidates = (
            session.query(generic_instance)
            .join(
                generic_instance_lineage,
                and_(
                    generic_instance_lineage.parent_instance_uid == source.uid,
                    generic_instance_lineage.child_instance_uid == generic_instance.uid,
                    generic_instance_lineage.relationship_type == parent_relationship_type,
                    generic_instance_lineage.is_deleted.is_(False),
                ),
            )
            .filter(
                generic_instance.template_uid == relation_template.uid,
                generic_instance.is_deleted.is_(False),
            )
            .all()
        )
        for candidate in candidates:
            parents = self.list_parents(
                session,
                child=candidate,
                relationship_type=child_relationship_type,
            )
            if parents:
                return candidate
        return None


def normalize_instance_payload(instance: generic_instance) -> dict[str, Any]:
    payload = dict(instance.json_addl or {})
    payload.setdefault("euid", instance.euid)
    payload.setdefault("name", instance.name)
    payload.setdefault(
        "created_at",
        instance.created_dt.isoformat().replace("+00:00", "Z")
        if instance.created_dt
        else utc_now_iso(),
    )
    payload.setdefault(
        "updated_at",
        instance.modified_dt.isoformat().replace("+00:00", "Z")
        if instance.modified_dt
        else payload["created_at"],
    )
    return payload



from __future__ import annotations

from dewey_service.app import create_app


def test_create_app_boots_with_injected_service(test_settings, fake_service) -> None:
    app = create_app(settings=test_settings, service=fake_service)
    assert app.title == "Dewey Artifact Service"
    assert app.state.observability is not None



from __future__ import annotations

from cli_core_yo import output
from cli_core_yo.app import run as run_cli

import dewey_service.cli as cli_module
import dewey_service.cli.db as db_cli


def _invoke(argv: list[str]) -> int:
    output._reset_console()
    return run_cli(cli_module.spec, argv)


def test_db_build_invokes_overlay(monkeypatch, capsys) -> None:
    calls: dict[str, int] = {"tapdb": 0, "dburl": 0, "seed": 0}

    class _Result:
        returncode = 0
        stdout = "ok"

    def _fake_tapdb(*args, **kwargs):
        calls["tapdb"] += 1
        return _Result()

    def _fake_db_url(*args, **kwargs):
        calls["dburl"] += 1
        return "postgresql+psycopg2://u:p@localhost:5439/dewey"

    def _fake_run(cmd, cwd=None, check=False):
        calls["seed"] += 1
        return _Result()

    monkeypatch.setattr(db_cli, "ensure_tapdb_version", lambda: "3.0.6")
    monkeypatch.setattr(db_cli, "run_tapdb_cli", _fake_tapdb)
    monkeypatch.setattr(db_cli, "export_database_url_for_target", _fake_db_url)
    monkeypatch.setattr(db_cli.subprocess, "run", _fake_run)

    exit_code = _invoke(["db", "build", "--target", "local"])
    capsys.readouterr()

    assert exit_code == 0
    assert calls["tapdb"] == 1
    assert calls["dburl"] == 1
    assert calls["seed"] == 1



from __future__ import annotations

import os
import shlex
import subprocess
from pathlib import Path

PROJECT_ROOT = Path(__file__).resolve().parents[1]
ACTIVATE_SCRIPT = PROJECT_ROOT / "activate"


def _write_executable(path: Path, content: str) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(content, encoding="utf-8")
    path.chmod(0o755)


def _build_fake_conda(tmp_path: Path) -> Path:
    conda_base = tmp_path / "fake-conda"
    conda_exe = conda_base / "bin" / "conda"
    env_bin = conda_base / "envs" / "DEWEY-local" / "bin"
    scripts_dir = conda_base / "envs" / "DEWEY-local" / "scripts"
    scripts_dir.mkdir(parents=True, exist_ok=True)

    _write_executable(
        env_bin / "python",
        f"""#!/usr/bin/env bash
set -euo pipefail
if [[ "${{1:-}}" == "-c" ]]; then
  if [[ "${{2:-}}" == *'sysconfig.get_path("scripts")'* ]]; then
    printf '%s\\n' "{scripts_dir}"
  fi
  exit 0
fi
if [[ "${{1:-}}" == "-m" && "${{2:-}}" == "pip" && "${{3:-}}" == "show" ]]; then
  exit 0
fi
if [[ "${{1:-}}" == "-m" && "${{2:-}}" == "pip" && "${{3:-}}" == "install" ]]; then
  exit 0
fi
exit 0
""",
    )

    _write_executable(
        conda_exe,
        f"""#!/usr/bin/env bash
set -euo pipefail
if [[ "${{1:-}}" == "info" && "${{2:-}}" == "--base" ]]; then
  printf '%s\\n' "{conda_base}"
  exit 0
fi
if [[ "${{1:-}}" == "info" && "${{2:-}}" == "--envs" ]]; then
  if [[ "${{FAKE_DEWEY_ENV_PRESENT:-1}}" == "1" ]]; then
    printf '# conda environments:\\n#\\nbase * {conda_base}\\nDEWEY-local {conda_base}/envs/DEWEY-local\\n'
  else
    printf '# conda environments:\\n#\\nbase * {conda_base}\\n'
  fi
  exit 0
fi
if [[ "${{1:-}}" == "env" && "${{2:-}}" == "create" ]]; then
  if [[ "${{FAKE_CONDA_ENV_CREATE_FAIL:-0}}" == "1" ]]; then
    exit 1
  fi
  exit 0
fi
echo "unexpected conda executable call: $*" >&2
exit 1
""",
    )

    conda_sh = conda_base / "etc" / "profile.d" / "conda.sh"
    conda_sh.parent.mkdir(parents=True, exist_ok=True)
    conda_sh.write_text(
        f"""conda() {{
  if [[ "${{1:-}}" == "activate" && "${{2:-}}" == "DEWEY-local" ]]; then
    if [[ -n "${{FAKE_CONDA_CALL_LOG:-}}" ]]; then
      printf 'activate\\n' >> "${{FAKE_CONDA_CALL_LOG}}"
    fi
    if [[ "${{FAKE_CONDA_ACTIVATE_FAIL:-0}}" == "1" ]]; then
      return 1
    fi
    export CONDA_DEFAULT_ENV="DEWEY-local"
    export CONDA_PREFIX="{conda_base}/envs/DEWEY-local"
    return 0
  fi
  command "{conda_exe}" "$@"
}}
""",
        encoding="utf-8",
    )

    return conda_base


def _source_activate(env: dict[str, str]) -> subprocess.CompletedProcess[str]:
    command = f"source {shlex.quote(str(ACTIVATE_SCRIPT))}"
    return subprocess.run(
        ["/bin/bash", "--noprofile", "--norc", "-c", command],
        cwd=PROJECT_ROOT,
        env=env,
        text=True,
        capture_output=True,
    )


def test_activate_requires_conda_on_path(tmp_path: Path) -> None:
    empty_bin = tmp_path / "bin"
    empty_bin.mkdir()

    env = os.environ.copy()
    env["PATH"] = "/usr/bin:/bin"
    env.pop("CONDA_DEFAULT_ENV", None)
    env.pop("CONDA_PREFIX", None)

    result = _source_activate(env)

    assert result.returncode == 1
    assert "Conda is required but was not found on PATH." in result.stderr


def test_activate_hardfails_when_conda_env_creation_fails(tmp_path: Path) -> None:
    conda_base = _build_fake_conda(tmp_path)

    env = os.environ.copy()
    env["PATH"] = f"{conda_base / 'bin'}:/usr/bin:/bin"
    env["FAKE_DEWEY_ENV_PRESENT"] = "0"
    env["FAKE_CONDA_ENV_CREATE_FAIL"] = "1"
    env.pop("CONDA_DEFAULT_ENV", None)
    env.pop("CONDA_PREFIX", None)

    result = _source_activate(env)

    assert result.returncode == 1
    assert "Failed to create conda environment from dewey_env.yaml." in result.stderr
    assert "Installing dewey CLI..." not in result.stdout


def test_activate_hardfails_when_conda_activation_fails(tmp_path: Path) -> None:
    conda_base = _build_fake_conda(tmp_path)

    env = os.environ.copy()
    env["PATH"] = f"{conda_base / 'bin'}:/usr/bin:/bin"
    env["FAKE_CONDA_ACTIVATE_FAIL"] = "1"
    env.pop("CONDA_DEFAULT_ENV", None)
    env.pop("CONDA_PREFIX", None)

    result = _source_activate(env)

    assert result.returncode == 1
    assert "Failed to activate conda environment: DEWEY-local" in result.stderr
    assert "Installing dewey CLI..." not in result.stdout


def test_activate_accepts_preloaded_dewey_conda_env(tmp_path: Path) -> None:
    conda_base = _build_fake_conda(tmp_path)
    call_log = tmp_path / "conda-calls.log"

    env = os.environ.copy()
    env["PATH"] = f"{conda_base / 'bin'}:/usr/bin:/bin"
    env["CONDA_DEFAULT_ENV"] = "DEWEY-local"
    env["CONDA_PREFIX"] = str(conda_base / "envs" / "DEWEY-local")
    env["FAKE_CONDA_CALL_LOG"] = str(call_log)

    result = _source_activate(env)

    assert result.returncode == 0
    assert "Conda environment already active: DEWEY-local" in result.stdout
    assert not call_log.exists()



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import pytest
import typer

import dewey_service.cli.db as db_cli
import dewey_service.cli.server as server_cli
from dewey_service.settings import Settings


def _settings() -> Settings:
    return Settings(
        cognito_domain="https://auth.example.com",
        cognito_app_client_id="client-1",
        cognito_redirect_uri="https://localhost:8914/auth/callback",
        cognito_logout_url="https://localhost:8914/login",
    )


def _proc(returncode: int = 0, stdout: str = "", stderr: str = "") -> SimpleNamespace:
    return SimpleNamespace(returncode=returncode, stdout=stdout, stderr=stderr)


def test_server_state_and_path_helpers(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    ctx = SimpleNamespace(xdg_paths=SimpleNamespace(state=tmp_path))
    monkeypatch.setattr("cli_core_yo.runtime.get_context", lambda: ctx)

    assert server_cli._state_dir() == tmp_path
    assert server_cli._log_dir() == tmp_path / "logs"
    assert server_cli._pid_file() == tmp_path / "server.pid"

    server_cli._ensure_runtime_dirs()
    assert (tmp_path / "logs").exists()


def test_server_load_settings_and_uri_validation(monkeypatch: pytest.MonkeyPatch, capsys) -> None:
    clear_calls: list[str] = []
    monkeypatch.setattr(server_cli, "clear_settings_cache", lambda: clear_calls.append("clear"))
    monkeypatch.setattr(server_cli, "get_settings", lambda: _settings())
    assert server_cli._load_settings().cognito_app_client_id == "client-1"
    assert clear_calls == ["clear"]

    monkeypatch.setattr(server_cli, "_load_settings", lambda: (_ for _ in ()).throw(ValueError("bad config")))
    with pytest.raises(typer.Exit) as exc:
        server_cli._validate_cognito_uris_for_port(port=8914, host="localhost")
    assert exc.value.exit_code == 1

    monkeypatch.setattr(server_cli, "_load_settings", _settings)
    monkeypatch.setattr(server_cli, "validate_uri_list_ports", lambda **kwargs: ["port mismatch"])
    server_cli._validate_cognito_uris_for_port(port=8914, host="localhost")
    assert "port mismatch" in capsys.readouterr().out

    monkeypatch.setattr(server_cli, "validate_uri_list_ports", lambda **kwargs: [])
    server_cli._validate_cognito_uris_for_port(port=8914, host="localhost")


def test_server_port_host_and_status_resolution(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("DEWEY_PORT", raising=False)
    assert server_cli._resolve_port(8914) == 8914

    monkeypatch.setenv("DEWEY_PORT", "9000")
    assert server_cli._resolve_port(8914) == 9000

    monkeypatch.setenv("DEWEY_PORT", "bad")
    with pytest.raises(typer.BadParameter):
        server_cli._resolve_port(8914)

    monkeypatch.setenv("DEWEY_HOST", "0.0.0.0")
    monkeypatch.setenv("DEWEY_PORT", "9001")
    assert server_cli._resolve_host("127.0.0.1") == "0.0.0.0"
    assert server_cli._status_bind() == ("localhost", "9001")

    monkeypatch.delenv("DEWEY_HOST", raising=False)
    monkeypatch.delenv("DEWEY_PORT", raising=False)
    monkeypatch.setattr(server_cli, "_load_settings", _settings)
    assert server_cli._status_bind() == ("localhost", "8914")

    monkeypatch.setattr(server_cli, "_load_settings", lambda: (_ for _ in ()).throw(ValueError("invalid")))
    assert server_cli._status_bind() == ("unknown", "unknown")


def test_start_server_branches(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(server_cli, "_ensure_runtime_dirs", lambda: None)
    monkeypatch.setattr(server_cli, "_pid_file", lambda: tmp_path / "server.pid")
    monkeypatch.setattr(server_cli, "_log_dir", lambda: tmp_path)

    cert = tmp_path / "cert.pem"
    key = tmp_path / "key.pem"
    monkeypatch.setattr(server_cli, "CERT_FILE", cert)
    monkeypatch.setattr(server_cli, "KEY_FILE", key)

    with pytest.raises(typer.BadParameter, match="HTTPS certs are missing"):
        server_cli._start_server(host="0.0.0.0", port=8914, reload=False, background=True)

    cert.write_text("cert", encoding="utf-8")
    key.write_text("key", encoding="utf-8")

    monkeypatch.setattr(server_cli, "read_pid", lambda path: 321)
    server_cli._start_server(host="0.0.0.0", port=8914, reload=False, background=True)

    monkeypatch.setattr(server_cli, "read_pid", lambda path: None)
    monkeypatch.setattr(server_cli, "new_log_path", lambda path: tmp_path / "server.log")
    monkeypatch.setattr(server_cli.time, "sleep", lambda seconds: None)
    monkeypatch.setattr(server_cli.shutil, "which", lambda name: "/usr/bin/python")

    class FailedProc:
        pid = 999

        @staticmethod
        def poll():
            return 1

    monkeypatch.setattr(server_cli.subprocess, "Popen", lambda *args, **kwargs: FailedProc())
    with pytest.raises(typer.Exit) as exc:
        server_cli._start_server(host="0.0.0.0", port=8914, reload=True, background=True)
    assert exc.value.exit_code == 1

    writes: list[int] = []

    class RunningProc:
        pid = 555

        @staticmethod
        def poll():
            return None

    monkeypatch.setattr(server_cli.subprocess, "Popen", lambda *args, **kwargs: RunningProc())
    monkeypatch.setattr(server_cli, "write_pid", lambda path, pid: writes.append(pid))
    server_cli._start_server(host="0.0.0.0", port=8914, reload=False, background=True)
    assert writes == [555]

    uvicorn_calls: list[dict[str, object]] = []
    monkeypatch.setattr(server_cli.uvicorn, "run", lambda **kwargs: uvicorn_calls.append(kwargs))
    server_cli._start_server(host="127.0.0.1", port=8915, reload=True, background=False)
    assert uvicorn_calls[0]["host"] == "127.0.0.1"
    assert uvicorn_calls[0]["port"] == 8915


def test_stop_server_and_logs_branches(monkeypatch: pytest.MonkeyPatch, tmp_path: Path, capsys) -> None:
    monkeypatch.setattr(server_cli, "_pid_file", lambda: tmp_path / "server.pid")
    monkeypatch.setattr(server_cli, "stop_pid", lambda path: (True, "Server stopped"))
    server_cli._stop_server()
    assert "Server stopped" in capsys.readouterr().out

    monkeypatch.setattr(server_cli, "stop_pid", lambda path: (False, "Permission denied stopping PID 1"))
    with pytest.raises(typer.Exit) as exc:
        server_cli._stop_server()
    assert exc.value.exit_code == 1

    monkeypatch.setattr(server_cli, "stop_pid", lambda path: (False, "No server running"))
    server_cli._stop_server()
    assert "No server running" in capsys.readouterr().out

    monkeypatch.setattr(server_cli, "_ensure_runtime_dirs", lambda: None)
    monkeypatch.setattr(server_cli, "_log_dir", lambda: tmp_path)
    monkeypatch.setattr(server_cli, "list_logs", lambda path: [])
    server_cli.logs(all_logs=True)
    assert "No log files found" in capsys.readouterr().out

    log_file = tmp_path / "server_1.log"
    log_file.write_text("hello", encoding="utf-8")
    monkeypatch.setattr(server_cli, "list_logs", lambda path: [log_file])
    server_cli.logs(all_logs=True)
    assert "server_1.log" in capsys.readouterr().out

    monkeypatch.setattr(server_cli, "latest_log", lambda path: None)
    server_cli.logs(all_logs=False)
    assert "No log file found" in capsys.readouterr().out

    monkeypatch.setattr(server_cli, "latest_log", lambda path: log_file)
    followed: list[tuple[Path, int]] = []
    monkeypatch.setattr(server_cli, "tail_follow", lambda path, lines=50: followed.append((path, lines)))
    server_cli.logs(lines=25, all_logs=False)
    assert followed == [(log_file, 25)]

    def raise_keyboard(path, lines=50):
        raise KeyboardInterrupt

    monkeypatch.setattr(server_cli, "tail_follow", raise_keyboard)
    server_cli.logs(lines=10, all_logs=False)


def test_server_command_wrappers(monkeypatch: pytest.MonkeyPatch) -> None:
    start_calls: list[dict[str, object]] = []
    monkeypatch.setenv("DEWEY_HOST", "127.0.0.1")
    monkeypatch.setenv("DEWEY_PORT", "9002")
    monkeypatch.setattr(server_cli, "_validate_cognito_uris_for_port", lambda **kwargs: start_calls.append({"validated": kwargs}))
    monkeypatch.setattr(server_cli, "_start_server", lambda **kwargs: start_calls.append(kwargs))

    server_cli.start(reload=True, background=False)
    assert start_calls[0]["validated"] == {"port": 9002, "host": "127.0.0.1"}
    assert start_calls[1] == {"host": "127.0.0.1", "port": 9002, "reload": True, "background": False}

    stop_calls: list[str] = []
    monkeypatch.setattr(server_cli, "_stop_server", lambda: stop_calls.append("stop"))
    server_cli.stop()
    assert stop_calls == ["stop"]


def test_db_cli_error_branches(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(db_cli, "ensure_tapdb_version", lambda: "3.0.6")

    with pytest.raises(typer.Exit) as exc:
        db_cli.build(target="aurora", cluster="", profile="lsmc", region="us-west-2", namespace="dewey")
    assert exc.value.exit_code == 1

    monkeypatch.setattr(
        db_cli,
        "run_tapdb_cli",
        lambda *args, **kwargs: _proc(returncode=0, stdout="ok"),
    )
    monkeypatch.setattr(
        db_cli,
        "export_database_url_for_target",
        lambda **kwargs: "postgresql+psycopg2://dewey@localhost:5432/dewey",
    )

    def fail_seed(cmd, cwd=None, check=False):
        raise db_cli.subprocess.CalledProcessError(returncode=7, cmd=cmd)

    monkeypatch.setattr(db_cli.subprocess, "run", fail_seed)
    with pytest.raises(typer.Exit) as exc:
        db_cli.build(target="local", cluster="", profile="lsmc", region="us-west-2", namespace="dewey")
    assert exc.value.exit_code == 1

    with pytest.raises(typer.Exit) as exc:
        db_cli.seed()
    assert exc.value.exit_code == 7

    monkeypatch.setattr(db_cli.typer, "confirm", lambda message: False)
    with pytest.raises(typer.Exit) as exc:
        db_cli.reset(force=False)
    assert exc.value.exit_code == 0

    monkeypatch.setattr(db_cli.typer, "confirm", lambda message: True)
    monkeypatch.setattr(
        db_cli,
        "run_tapdb_cli",
        lambda *args, **kwargs: (_ for _ in ()).throw(db_cli.TapDBRuntimeError("delete failed")),
    )
    with pytest.raises(typer.Exit) as exc:
        db_cli.reset(
            force=True,
            target="local",
            cluster="",
            profile="lsmc",
            region="us-west-2",
            namespace="dewey",
        )
    assert exc.value.exit_code == 1


def test_db_cli_reset_success_path(monkeypatch: pytest.MonkeyPatch) -> None:
    calls: list[tuple[str, object]] = []
    monkeypatch.setattr(
        db_cli,
        "run_tapdb_cli",
        lambda args, **kwargs: calls.append(("delete", args)) or _proc(returncode=0, stdout="deleted"),
    )
    monkeypatch.setattr(
        db_cli,
        "build",
        lambda **kwargs: calls.append(("build", kwargs)),
    )

    db_cli.reset(force=True, target="local", cluster="cluster-1", profile="lsmc", region="us-west-2", namespace="dewey")

    assert calls == [
        ("delete", ["db", "delete", "dev", "--force"]),
        (
            "build",
            {
                "target": "local",
                "cluster": "cluster-1",
                "profile": "lsmc",
                "region": "us-west-2",
                "namespace": "dewey",
            },
        ),
    ]



from __future__ import annotations


def test_artifact_detail_and_storage_routes(client) -> None:
    created = client.post(
        "/api/v1/artifacts",
        headers={"Authorization": "Bearer token-123", "Idempotency-Key": "idem-gap-art-1"},
        json={
            "artifact_type": "vcf",
            "storage_backend": "s3",
            "bucket": "bucket-1",
            "key": "variants/sample.vcf.gz",
            "metadata": {"source": "coverage-gap"},
        },
    )
    assert created.status_code == 200
    artifact_euid = created.json()["artifact_euid"]

    fetched = client.get(
        f"/api/v1/artifacts/{artifact_euid}",
        headers={"Authorization": "Bearer token-123"},
    )
    assert fetched.status_code == 200
    assert fetched.json()["artifact_euid"] == artifact_euid
    assert fetched.json()["metadata"] == {"source": "coverage-gap"}

    verified = client.post(
        f"/api/v1/artifacts/{artifact_euid}/storage/verify",
        headers={"Authorization": "Bearer token-123", "Idempotency-Key": "idem-gap-art-verify"},
    )
    assert verified.status_code == 200
    assert verified.json()["storage_status"] == "verified"
    assert verified.json()["storage_verified_at"].startswith("2026-03-10T00:00:00")

    locked = client.post(
        f"/api/v1/artifacts/{artifact_euid}/storage/lock",
        headers={"Authorization": "Bearer token-123", "Idempotency-Key": "idem-gap-art-lock"},
        json={"mode": "COMPLIANCE", "retain_until": "2028-01-01T00:00:00Z"},
    )
    assert locked.status_code == 200
    assert locked.json()["retention_mode"] == "COMPLIANCE"
    assert locked.json()["retain_until"] == "2028-01-01T00:00:00Z"


def test_upload_session_complete_route_round_trip(client) -> None:
    created = client.post(
        "/api/v1/artifacts/upload-sessions",
        headers={"Authorization": "Bearer token-123", "Idempotency-Key": "idem-gap-upload-create"},
        json={
            "artifact_type": "report",
            "original_filename": "coverage-gap.pdf",
            "content_type": "application/pdf",
            "producer_system": "atlas",
            "producer_object_euid": "REL-coverage-gap",
        },
    )
    assert created.status_code == 200
    upload_token = created.json()["upload_token"]

    completed = client.post(
        f"/api/v1/artifacts/upload-sessions/{upload_token}/complete",
        headers={"Authorization": "Bearer token-123", "Idempotency-Key": "idem-gap-upload-complete"},
        json={"checksums": {"sha256": "abc123"}, "metadata": {"stage": "gap"}},
    )
    assert completed.status_code == 200
    payload = completed.json()
    assert payload["status_code"] == 201
    assert payload["artifact_type"] == "report"
    assert payload["original_filename"] == "coverage-gap.pdf"
    assert payload["import_mode"] == "upload"


def test_share_reference_lookup_and_external_object_relation_routes(client) -> None:
    artifact = client.post(
        "/api/v1/artifacts",
        headers={"Authorization": "Bearer token-123", "Idempotency-Key": "idem-gap-share-art"},
        json={
            "artifact_type": "json",
            "storage_backend": "s3",
            "bucket": "bucket-2",
            "key": "reports/share-gap.json",
        },
    ).json()

    share = client.post(
        "/api/v1/share-references",
        headers={"Authorization": "Bearer token-123", "Idempotency-Key": "idem-gap-share-ref"},
        json={
            "target_type": "artifact",
            "target_euid": artifact["artifact_euid"],
            "purpose": "download",
            "scope": "external",
            "transport": "presigned_s3",
        },
    )
    assert share.status_code == 200
    share_reference_euid = share.json()["share_reference_euid"]

    fetched_share = client.get(
        f"/api/v1/share-references/{share_reference_euid}",
        headers={"Authorization": "Bearer token-123"},
    )
    assert fetched_share.status_code == 200
    assert fetched_share.json()["share_reference_euid"] == share_reference_euid

    listed_share_refs = client.get(
        f"/api/v1/artifacts/{artifact['artifact_euid']}/share-references",
        headers={"Authorization": "Bearer token-123"},
    )
    assert listed_share_refs.status_code == 200
    assert listed_share_refs.json()["total"] == 1

    external_object = client.post(
        "/api/v1/external-objects",
        headers={"Authorization": "Bearer token-123", "Idempotency-Key": "idem-gap-external-object"},
        json={
            "external_system": "atlas",
            "external_object_type": "document",
            "external_object_id": "DOC-GAP-1",
            "external_uri": "https://atlas.example/documents/DOC-GAP-1",
        },
    )
    assert external_object.status_code == 200
    external_object_euid = external_object.json()["external_object_euid"]

    relation = client.post(
        "/api/v1/external-object-relations",
        headers={"Authorization": "Bearer token-123", "Idempotency-Key": "idem-gap-external-relation"},
        json={
            "target_type": "artifact",
            "target_euid": artifact["artifact_euid"],
            "external_object_euid": external_object_euid,
            "relation_type": "source_record",
        },
    )
    assert relation.status_code == 200

    listed_relations = client.get(
        f"/api/v1/artifact/{artifact['artifact_euid']}/external-object-relations",
        headers={"Authorization": "Bearer token-123"},
    )
    assert listed_relations.status_code == 200
    assert listed_relations.json()["total"] == 1
